# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Documents the binary format of Capabilities.

A capability is not an image metadata section, but rather the content.
Image metadata is appended to the end of a capability file.
"""

from ..metadata import MetadataSection
import struct


class Capability:
    # /// <summary>
    # /// default case with no capability.
    # /// NoCapability indicates  no further capability is following.
    # /// </summary>
    NoCapability = 0x0

    # /// <summary>
    # /// allow test key in secure boot for development purpose.
    # /// </summary>
    AllowTestKey = 0x1

    # /// <summary>
    # /// enable Pluton M4 debug.
    # /// </summary>
    EnablePlutonDebug = 0x2

    # /// <summary>
    # /// enable Hlos Core debug.
    # /// </summary>
    EnableHlosCoreDebug = 0x3

    # /// <summary>
    # /// enable N9 debug.
    # /// </summary>
    EnableN9Debug = 0x4

    # /// <summary>
    # /// enable Hlos Core console.
    # /// </summary>
    EnableHlosCoreConsole = 0x8

    # /// <summary>
    # /// Enable SLT Loader from 1BL
    # /// </summary>
    EnableSltLoader = 0x9

    # /// <summary>
    # /// enable System Software Development.
    # /// </summary>
    EnableSystemSoftwareDevelopment = 0xA

    # /// <summary>
    # /// enable Customer App Development.
    # /// </summary>
    EnableCustomerAppDevelopment = 0xB

    # /// <summary>
    # /// Enable entering low-level RF Test Mode.
    # /// </summary>
    EnableRfTestMode = 0xC

    # /// <summary>
    # /// Unlock GatewayD host communication
    # /// </summary>
    UnlockGatewayD = 0xD

    def __init__(self, capability: bytearray):
        self.capability = capability

    def __str__(self):
        return self._capability_to_name(self.capability)

    def _capability_to_name(self, capability):
        if capability == Capability.NoCapability:
            return "NoCapability"
        elif capability == Capability.AllowTestKey:
            return "AllowTestKey"
        elif capability == Capability.EnablePlutonDebug:
            return "EnablePlutonDebug"
        elif capability == Capability.EnableHlosCoreDebug:
            return "EnableHlosCoreDebug"
        elif capability == Capability.EnableN9Debug:
            return "EnableN9Debug"
        elif capability == Capability.EnableHlosCoreConsole:
            return "EnableHlosCoreConsole"
        elif capability == Capability.EnableSltLoader:
            return "EnableSltLoader"
        elif capability == Capability.EnableSystemSoftwareDevelopment:
            return "EnableSystemSoftwareDevelopment"
        elif capability == Capability.EnableCustomerAppDevelopment:
            return "EnableCustomerAppDevelopment"
        elif capability == Capability.EnableRfTestMode:
            return "EnableRfTestMode"
        elif capability == Capability.UnlockGatewayD:
            return "UnlockGatewayD"
        return "UnknownCapability"


class CapabilitiesSection(MetadataSection):
    """
    A capability file contains:
    # uint32_t magic 0x5CFD5CFD
    # uint32_t version
    # uint32_t data_length
    # bytes[64] device_id
    # bytes[64] capabilities
    """

    def __init__(self, data):
        super().__init__("Capabilities", None, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.magic_value = 0
            self.version = 0
            self.data_length = 0
            self.device_id = [0] * 64
            self.capabilities_raw = [0] * 64

    @staticmethod
    def size():
        """
        Returns the size of the capabilities section in bytes.
        """
        return 136

    @property
    def capabilities(self):
        """
        Returns a list of capabilities.
        """
        return [capability for capability in self.capabilities_raw if capability != 0]

    @capabilities.setter
    def capabilities(self, value: bytearray):
        """
        Sets the capabilities using the provided bytearray.
        Each index in the bytearray represents a capability.
        """
        for idx, v in enumerate(value):
            self.capabilities_raw[idx] = v

    def deserialize(self, data: bytearray):
        """
        Given a bytearray, deserialize the data into properties
        """
        self.magic_value = struct.unpack_from("<I", data, 0)[0]
        self.version = struct.unpack_from("<I", data, 4)[0]
        self.data_length = struct.unpack_from("<I", data, 8)[0]
        self.device_id_raw = bytearray(data[12:76])
        self.device_id = self.device_id_raw.hex()
        self.capabilities_raw = data[76:140]

        if self.magic_value != 0x5CFD5CFD:
            raise Exception("CapabilityStructure magic is not correct")

    def serialize(self) -> bytes:
        """
        Serializes the data into a bytearray.
        """
        data = bytearray(self.size())
        struct.pack_into("<I", data, 0, self.magic_value)
        struct.pack_into("<I", data, 4, self.version)
        struct.pack_into("<I", data, 8, self.data_length)
        data[12:76] = self.device_id_raw
        data[76:140] = self.capabilities_raw
        return data

    def __str__(self):
        out_lines = []
        out_lines.append("Capability Data:")
        out_lines.append(f"\tMagic: {self.magic_value}")
        out_lines.append(f"\tMetadata Version: {self.version}")
        out_lines.append(f"\tLength: {self.data_length}")
        out_lines.append(f"\tDevice ID: {self.device_id}")
        out_lines.append("\tCapabilities:")
        for capability in self.capabilities:
            cap = Capability(capability)
            out_lines.append(f"\t\t{cap}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The Debug Section provides additional useful metadata about the image.
This includes things like build time, and image name.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct
from datetime import datetime, timezone


class DebugSection(MetadataSection):
    """
    The Debug Section provides additional useful metadata about the image:
    - The time the image was built.
    - The name of the image.
    """

    def __init__(self, data=None):
        super().__init__("Debug", MetadataSectionId.Debug, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.debug_time = 0
            self.image_name_raw = bytes([0] * 32)
            self.image_name = ""

    @staticmethod
    def size():
        """
        The size of the Debug Section is 8 bytes for the time, and 32 bytes for the image name.
        """
        return 8 + 32

    def deserialize(self, data):
        """
        Deserialize the Debug Section from the provided data.
        """
        self.debug_time = struct.unpack_from("<Q", data, 0)[0]
        self.image_name_raw = struct.unpack_from("<32s", data, 8)[0]
        self.image_name = str(self.image_name_raw, encoding="utf-8")
        self.image_name = self.image_name.strip("\0")

    def serialize(self):
        """
        Serialize the Debug Section to a byte array.
        """
        return struct.pack("<Q32s", self.debug_time, self.image_name_raw)

    def __str__(self):
        utc_orig = datetime.utcfromtimestamp(self.debug_time)
        LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo
        local = datetime.fromtimestamp(self.debug_time, LOCAL_TIMEZONE)

        out_lines = []
        out_lines.append(f"Name:              {self.image_name}")
        out_lines.append(f"Built On (UTC):    {utc_orig.strftime('%d/%m/%Y %H:%M:%S')}")
        out_lines.append(f"Built On (Local):  {local.strftime('%d/%m/%Y %H:%M:%S')}")

        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The identity section contains information about the image, such as the image type, version, and size.
All images must have an identity metadata section.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
from ..guid_utils import format_guid_as_string, to_guid
import struct


class ImageType:
    """
    Image type enum.
    """

    # /// <summary>
    # /// Invalid.
    # /// </summary>
    Invalid = 0

    # /// <summary>
    # /// Better known as 1BL; code executed by the ROM on Pluton.
    # /// </summary>
    OneBL = 1

    # /// <summary>
    # /// Pluton runtime.
    # /// </summary>
    PlutonRuntime = 2

    # /// <summary>
    # /// Vendor-specific WiFi firmware.
    # /// </summary>
    WifiFirmware = 3

    # /// <summary>
    # /// Secure World runtime, a.k.a. security monitor.
    # /// </summary>
    SecurityMonitor = 4

    # /// <summary>
    # /// Normal World loader.
    # /// </summary>
    NormalWorldLoader = 5

    # /// <summary>
    # /// Device Tree, used by Normal World OS.
    # /// </summary>
    NormalWorldDTB = 6

    # /// <summary>
    # /// Normal World OS kernel.
    # /// </summary>
    NormalWorldKernel = 7

    # /// <summary>
    # /// Normal World root file system.
    # /// </summary>
    RootFs = 8

    # /// <summary>
    # /// Normal World, services written and managed by Microsoft.
    # /// </summary>
    Services = 9

    # /// <summary>
    # /// Normal World, user-mode apps.
    # /// </summary>
    Applications = 10

    # /// <summary>
    # /// Configuration data for firmware (Pluton, HlosCore).
    # /// </summary>
    FirmwareConfig = 13

    # /// <summary>
    # /// Boot Manifest (secure flash storage)
    # /// </summary>
    BootManifest = 16

    # /// <summary>
    # /// Normal World File System
    # /// </summary>
    NormalWorldFileSystem = 17

    # ///<summary>
    # /// The TrustedKeystore image
    # ///</summary>
    TrustedKeystore = 19

    # ///<summary>
    # /// An image containing policies for the device.
    # ///</summary>
    Policy = 20

    # /// <summary>
    # /// An image containing board configuration for each customer system.
    # /// </summary>
    CustomerBoardConfig = 21

    # /// <summary>
    # /// An image containing the certificates needed by the OTA update process.
    # /// </summary>
    UpdateCertStore = 22

    # /// <summary>
    # /// A manifest describing the update of our trusted keys and certificate stores.
    # /// </summary>
    BaseSystemUpdateManifest = 23

    # /// <summary>
    # /// A manifest describing the update of our system software.
    # /// </summary>
    FirmwareUpdateManifest = 24

    # /// <summary>
    # /// A manifest describing the update of 3rd-party images.
    # /// </summary>
    CustomerUpdateManifest = 25

    # /// <summary>
    # /// A manifest describing system software for recovery.
    # /// </summary>
    RecoveryManifest = 26

    # /// <summary>
    # /// A manifest set that contains 1 or more manifests and flags concerning those manifests.
    # /// </summary>
    ManifestSet = 27

    # /// <summary>
    # /// Sentinel value indicating an unspecified image type.
    # /// </summary>
    Other = 28

    @staticmethod
    def get_name(image_type) -> str:
        """
        Returns the name of the provided image type.
        """
        if image_type == ImageType.Invalid:
            return "Invalid"
        elif image_type == ImageType.OneBL:
            return "OneBL"
        elif image_type == ImageType.PlutonRuntime:
            return "PlutonRuntime"
        elif image_type == ImageType.WifiFirmware:
            return "WifiFirmware"
        elif image_type == ImageType.SecurityMonitor:
            return "SecurityMonitor"
        elif image_type == ImageType.NormalWorldLoader:
            return "NormalWorldLoader"
        elif image_type == ImageType.NormalWorldDTB:
            return "NormalWorldDTB"
        elif image_type == ImageType.NormalWorldKernel:
            return "NormalWorldKernel"
        elif image_type == ImageType.RootFs:
            return "RootFs"
        elif image_type == ImageType.Services:
            return "Services"
        elif image_type == ImageType.Applications:
            return "Applications"
        elif image_type == ImageType.FirmwareConfig:
            return "FirmwareConfig"
        elif image_type == ImageType.BootManifest:
            return "BootManifest"
        elif image_type == ImageType.NormalWorldFileSystem:
            return "NormalWorldFileSystem"
        elif image_type == ImageType.TrustedKeystore:
            return "TrustedKeystore"
        elif image_type == ImageType.Policy:
            return "Policy"
        elif image_type == ImageType.CustomerBoardConfig:
            return "CustomerBoardConfig"
        elif image_type == ImageType.UpdateCertStore:
            return "UpdateCertStore"
        elif image_type == ImageType.BaseSystemUpdateManifest:
            return "BaseSystemUpdateManifest"
        elif image_type == ImageType.FirmwareUpdateManifest:
            return "FirmwareUpdateManifest"
        elif image_type == ImageType.CustomerUpdateManifest:
            return "CustomerUpdateManifest"
        elif image_type == ImageType.RecoveryManifest:
            return "RecoveryManifest"
        elif image_type == ImageType.ManifestSet:
            return "ManifestSet"
        elif image_type == ImageType.Other:
            return "Other"
        else:
            return "Unknown"


class IdentitySection(MetadataSection):
    """
    Identity section of the metadata.
    Contains information about the image included the image type, uuid and component id.
    """

    # identity
    # ushort image_tyoe
    # ushort reserved
    # byte[16] uuid
    # byte[16] component_id

    def __init__(self, data=None):
        super().__init__("Identity", MetadataSectionId.Identity, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.image_type = ImageType.Invalid
            self.reserved = 0
            self.image_uid_raw = bytes([0] * 16)
            self.component_uid_raw = bytes([0] * 16)

    @staticmethod
    def size():
        """
        Returns the size of the identity section in bytes.
        """
        return 36

    @property
    def image_uid(self):
        """
        Returns the image uuid as a string.
        """
        return format_guid_as_string(to_guid(self.image_uid_raw))

    @property
    def component_uid(self):
        """
        Returns the component uuid as a string.
        """
        return format_guid_as_string(to_guid(self.component_uid_raw))

    # TODO implement setters

    def deserialize(self, data):
        """
        Deserializes the identity section from the provided data.
        """
        self.image_type = struct.unpack_from("<H", data, 0)[0]
        self.reserved = struct.unpack_from("<H", data, 2)[0]
        self.component_uid_raw = struct.unpack_from("<16s", data, 4)[0]
        self.image_uid_raw = struct.unpack_from("<16s", data, 20)[0]

    def serialize(self):
        """
        Serializes the identity section into bytes.
        """
        data = bytearray(IdentitySection.size())
        struct.pack_into("<H", data, 0, self.image_type)
        struct.pack_into("<H", data, 2, self.reserved)
        struct.pack_into("<16s", data, 4, self.component_uid_raw)
        struct.pack_into("<16s", data, 20, self.image_uid_raw)
        return data

    def __str__(self):
        out_lines = []
        out_lines.append(f"Image Type:        {ImageType.get_name(self.image_type)}")
        out_lines.append(f"Component UID:     {self.component_uid}")
        out_lines.append(f"Image UID:         {self.image_uid}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
A class that deserialises/parses image metadata given an Azure Sphere image.
"""

import struct
from .metadata_sections.utils import class_from_section_id
from .metadata_sections.internal.onebl import OneBlInformationSection
from .metadata_headers.image_metadata import ImageMetadataHeader
from .metadata_headers.image_metadata_section import ImageMetadataSectionHeader


class ImageMetadata:
    """
    Image metadata contains a header, followed by a number of sections.
    """

    def __init__(self, data):
        self.sections = []
        self.section_headers = []
        self.header = None
        self.onebl = None
        self.end_of_metadata = 0
        self.start_of_metadata = 0
        self.signature_size = 0
        self.raw_data = bytes()
        if data is not None:
            self.deserialize(data)

    def deserialize(self, byte_data):
        """
        Image metadata is placed at the end of the file.
        """
        position = 0
        # The image signature is appended to the end of the metadata and is:
        # 64 bytes, if sha256 is used
        # 96 bytes, if sha384 is used
        possible_signature_sizes = [64, 96]
        for possible_signature_size in possible_signature_sizes:
            self.onebl = None  # reset onebl on loop as it can be set incorrectly
            minimum_valid_size = (
                possible_signature_size + 4 + ImageMetadataSectionHeader.size()
            )
            metadata_length_offset = possible_signature_size + 4
            onebl_metadata_length_offset = (
                possible_signature_size + 4 + OneBlInformationSection.size()
            )

            if len(byte_data) < minimum_valid_size:
                continue

            sig_area_size = possible_signature_size
            end_of_metadata = len(byte_data) - metadata_length_offset
            # length is stored in the four bytes before the signature
            target_length = struct.unpack_from("<I", byte_data, end_of_metadata)[0]

            if target_length <= 28:
                # This may be a 1BL only section
                if len(byte_data) < minimum_valid_size + OneBlInformationSection.size():
                    continue

                sig_area_size += OneBlInformationSection.size()

                position = (
                    len(byte_data)
                    - possible_signature_size
                    - OneBlInformationSection.size()
                )
                self.onebl = OneBlInformationSection(
                    byte_data[position : position + OneBlInformationSection.size()]
                )

                end_of_metadata = len(byte_data) - onebl_metadata_length_offset
                position = end_of_metadata
                target_length = struct.unpack_from("<I", byte_data, position)[0]

            if len(byte_data) < (sig_area_size + target_length):
                # Not enough space, fail
                continue

            if target_length < ImageMetadataHeader.size():
                # Length too small, fail
                continue

            header_position = len(byte_data) - sig_area_size - target_length
            self.start_of_metadata = header_position

            self.image_data = byte_data[0:header_position]

            header = ImageMetadataHeader(
                byte_data[
                    header_position : header_position + ImageMetadataHeader.size()
                ]
            )

            self.header = header

            if header.section_count == 0:
                continue

            section_addr = header_position + ImageMetadataHeader.size()

            for _ in range(0, header.section_count):
                if section_addr >= end_of_metadata:
                    continue

                section_header_data = byte_data[
                    section_addr : section_addr + ImageMetadataSectionHeader.size()
                ]
                section_header = ImageMetadataSectionHeader(section_header_data)

                self.section_headers += [section_header]

                if section_header.data_length == 0:
                    continue

                if (
                    section_addr
                    + ImageMetadataSectionHeader.size()
                    + section_header.data_length
                    > end_of_metadata
                ):
                    continue

                section_addr += ImageMetadataSectionHeader.size()

                self.sections += [
                    class_from_section_id(section_header.section_id)(
                        byte_data[
                            section_addr : section_addr + section_header.data_length
                        ]
                    )
                ]

                section_addr += section_header.data_length

            self.end_of_metadata = (
                end_of_metadata + 4
            )  # include the length in the metadata
            self.signature_size = possible_signature_size
            self.raw_data = byte_data[self.start_of_metadata : self.end_of_metadata]
            return
        raise Exception("Invalid metadata")

    def serialize(self):
        """
        Serialize the image metadata to a byte array.
        """
        header = ImageMetadataHeader()
        header.section_count = len(self.sections)

        out = header.serialize()

        for s in self.sections:
            sec_bytes = s.serialize()
            sec_header = ImageMetadataSectionHeader()
            sec_header.section_id = s.section_id
            sec_header.data_length = len(sec_bytes)
            out += sec_header.serialize()
            out += sec_bytes

        # Add the length of the metadata to the end
        # Include the length of the length field itself (4 bytes)
        out += struct.pack("<I", len(out) + 4)
        return out

    def add_section(self, section):
        """
        Add a section to the metadata.
        """
        self.sections += [section]

    def sections_by_name(self, name):
        """
        Returns a list of sections with the given name.
        """
        return [section for section in self.sections if section.name == name]

    def sections_by_id(self, section_id):
        """
        Returns a list of sections with the given id.
        """
        return [
            section for section in self.sections if section.section_id == section_id
        ]

    def remove_section(self, section):
        """
        Removes a section from the metadata.
        """
        self.sections = [s for s in self.sections if s != section]

    def str_dotnet(self) -> str:
        """
        Returns a string representation of the metadata using the same formatting errors as
        the dotnet tool.
        """
        out_lines = []
        # to match dotnet output exactly, each section header has \r\n at the end of it
        # each content line has \r\r\n at the end of it.
        # This is likely a bug in the dotnet code where the lines were split assuming \n
        # as the line ending when infact the line ending was \r\n
        # We will match it for testing purposes.
        for section in self.sections:
            out_lines.append(f"  Section: {section.name}\r\n")

            # the formatting error was not globally applied to all sections,
            # so we need to special case this one
            if section.name == "Required Offset":
                for line in str(section).split("\n"):
                    if line.strip() == "":
                        continue
                    out_lines.append(f"    {line}\r\n")
            else:
                for line in str(section).split("\n"):
                    if line.strip() == "":
                        continue
                    out_lines.append(f"    {line}\r\r\n")

        if self.onebl is not None:
            out_lines.append("  Section: 1BL Information\r\n")
            for line in str(self.onebl).split("\n"):
                if line.strip() == "":
                    continue
                out_lines.append(f"    {line}\r\r\n")

        out_str = "".join(out_lines)
        return out_str + "\r\n"

    def __str__(self):
        """
        Returns a correctly formatted string representation of the metadata.
        """
        out_lines = []
        for section in self.sections:
            out_lines.append(f"  Section: {section.name}")

            for line in str(section).split("\n"):
                if line.strip() == "":
                    continue
                out_lines.append(f"    {line}")

        if self.onebl is not None:
            out_lines.append("  Section: 1BL Information")
            for line in str(self.onebl).split("\n"):
                if line.strip() == "":
                    continue
                out_lines.append(f"    {line}")

        return "\r\n".join(out_lines) + "\r\n"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The image policy section indicates whether the image can be invalidated or sideloaded from normal world (OS).
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct
from typing import List


class ImagePolicyFlags:
    """
    Image policy flags enum.
    """

    # uint32_t flags
    NoFlags = 0
    InvalidateFromNormalWorld = 0x1 << 0
    SideloadFromNormalWorld = 0x1 << 1

    flags: int

    def __init__(self, data=None):
        if data is not None:
            self.deserialize(data)
        else:
            self.flags = self.NoFlags

    def get_flags_name(self) -> List[str]:
        """
        Converts the flags to a string array.
        """
        flags = []
        if self.flags & ImagePolicyFlags.InvalidateFromNormalWorld:
            flags += ["InvalidateFromNormalWorld"]
        if self.flags & ImagePolicyFlags.SideloadFromNormalWorld:
            flags += ["SideloadFromNormalWorld"]
        return flags

    @staticmethod
    def size():
        """
        Returns the size of the flags field in bytes.
        """
        return 4

    def deserialize(self, data):
        """
        Deserializes the flags from the given data.
        """
        self.flags = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        """
        Serializes the flags to bytes.
        """
        return struct.pack("<I", self.flags)

    def __str__(self):
        return "ImagePolicyFlags\n\tflags: %d" % (self.flags) + "\n"


class ImagePolicySection(MetadataSection):
    """
    The image policy section, if present, indicates whether the image can be invalidated or sideloaded from normal world (OS).
    """

    def __init__(self, data=None):
        super().__init__("Image Policy", MetadataSectionId.ImagePolicy, data)

        if data is not None:
            self.deserialize(data)
        else:
            self.flags = ImagePolicyFlags(None)

    @staticmethod
    def size():
        """
        Returns the size of the image policy section in bytes.
        """
        return ImagePolicyFlags.size()

    def deserialize(self, data):
        """
        Deserializes the image policy section from the given data.
        """
        self.flags = ImagePolicyFlags(data)

    def serialize(self):
        """
        Serializes the image policy section to bytes.
        """
        return self.flags.serialize()

    def __str__(self):
        invalid_from_nw = "InvalidateFromNormalWorld" in self.flags.get_flags_name()
        sideload_from_nw = "SideloadFromNormalWorld" in self.flags.get_flags_name()

        out_lines = []
        out_lines.append(f"Allow invalidation from Normal World: {invalid_from_nw}")
        out_lines.append(f"Allow sideloading from Normal World: {sideload_from_nw}")

        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Previously called "AppTestKeyResolver" in C# code, this class is used to sign applications for development.
"""

from hashlib import sha256
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives.asymmetric import ec, utils
from cryptography.exceptions import InvalidSignature
from pathlib import Path
import os

PARENT_DIR = Path(__file__).parent.parent


class AppDevelopmentSigner:
    """
    Uses the certificates in the ./certificates folder to sign applications for development.
    """

    def __init__(self):
        # current working theory is that tp_app_test_sign is for older image packages.
        with open(
            os.path.join(PARENT_DIR, "certificates", "tp_app_test_sign.cer"), "rb"
        ) as f:
            self.ctp_signing_cert = x509.load_pem_x509_certificate(f.read())

        # app_test_sign.pfx is for newer image packages and should be used by default.
        # code should be able to read signatures signed by either key,
        # but should only sign using app_test_sign.pfx
        with open(
            os.path.join(PARENT_DIR, "certificates", "app_test_sign.pfx"), "rb"
        ) as f:
            self.test_signing_cert = pkcs12.load_pkcs12(f.read(), None)

    def thumbprint(self):
        """
        returns the thumbprint of the "new"? app development certificate.
        """
        return self.test_signing_cert.cert.certificate.fingerprint(hashes.SHA1()).hex()

    def sign_data(self, data: bytes) -> bytes:
        """
        signs the provided data using the app development certificate
        use the "new"? certificate by default.

        returns the DER encoded signature.
        """
        return self.test_signing_cert.key.sign(
            sha256(data).digest(), ec.ECDSA(utils.Prehashed(hashes.SHA256()))
        )

    def verify(self, stored_signature: bytes, computed_hash: bytes) -> bool:
        """
        verifies the provided hash using the app development certificates.
        """
        certificates = [self.ctp_signing_cert, self.test_signing_cert.cert.certificate]

        for certificate in certificates:
            try:
                certificate.public_key().verify(
                    stored_signature,
                    computed_hash,
                    ec.ECDSA(utils.Prehashed(hashes.SHA256())),
                )
                return True
            except InvalidSignature as e:
                print(f"validation failed: {str(e)} asd")
                # if here, the signature failed to validate - continue through the list.
                pass

        return False



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Provides helper functions for working with Azure Sphere GUIDs.
Azure Sphere GUIDs have historically incorrectly ordered bytes.
"""


def format_guid_as_string(guid):
    """
    Formats a GUID as a string observing the same format as dotnet code.
    """
    return f"{guid[0:8]}-{guid[8:12]}-{guid[12:16]}-{guid[16:20]}-{guid[20:32]}"


def to_guid(guid_bytes: bytes):
    """
    Correctly orders a guid stored in an image package e.g. image/component uids.
    """
    guid = [0] * 16

    for i in range(0, 4):
        guid[i] = guid_bytes[3 - i]

    guid[4] = guid_bytes[5]
    guid[5] = guid_bytes[4]
    guid[6] = guid_bytes[7]
    guid[7] = guid_bytes[6]

    guid[8:] = guid_bytes[8:]

    return bytearray(guid).hex()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The signature section contains the certificate hash and signing type used to sign the image.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct


class SigningType:
    """
    Signing type enum.
    """

    Invalid = 0
    ECDsa256 = 1
    ECDsa384 = 2

    def __init__(self, data: bytes):
        if data is not None:
            self.deserialize(data)
        else:
            self.signing_type = 0

    def deserialize(self, data: bytes):
        """
        Deserializes the signing type from the given data.
        """
        self.signing_type = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        """
        Serializes the signing type to bytes.
        """
        return struct.pack("<I", self.signing_type)

    def get_name(self):
        """
        Get the name of the signing type.
        """
        if self.signing_type == SigningType.Invalid:
            return "Invalid"
        elif self.signing_type == SigningType.ECDsa256:
            return "ECDsa256"
        elif self.signing_type == SigningType.ECDsa384:
            return "ECDsa384"
        else:
            raise Exception("Unknown signing type")


class SignatureSection(MetadataSection):

    """
    The signature section contains the certificate hash and signing type used to sign the image.
    # byte[20] SigningCertThumbprint
    # uint32 signing type
    """

    def __init__(self, data=None):
        super().__init__("Signature", MetadataSectionId.Signature, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.signing_cert_thumbprint = b""
            self.signing_type = SigningType(None)

    @staticmethod
    def size():
        """
        Returns the size of the signature section in bytes.
        """
        return 24

    def deserialize(self, data):
        """
        Deserializes the signature section from the given data.
        """
        self.signing_cert_thumbprint = struct.unpack_from("<20s", data, 0)[0]
        self.signing_type = SigningType(data[20:])

    def serialize(self):
        return (
            struct.pack("<20s", self.signing_cert_thumbprint)
            + self.signing_type.serialize()
        )

    def __str__(self):
        cert = self.signing_cert_thumbprint
        out_lines = []
        out_lines.append(f"Signing Type:      {self.signing_type.get_name()}")
        out_lines.append(f"Cert:              {cert.hex()}")
        out_lines.append(f"Cert Thumbprint:   {cert.hex()}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
MetadataSection is the super class for all metadata sections.
It contains common properties (name, section_id, raw_data).
"""
from .metadata_id import MetadataSectionId


class MetadataSection:
    """
    MetadataSection is the super class for all metadata sections.
    """

    name: str
    section_id: MetadataSectionId
    raw_data: bytearray

    def __init__(self, name: str, section_id: MetadataSectionId, data: bytearray):
        self.name = name
        self.section_id = section_id
        # preserve a copy of the original data buf.
        self.raw_data = data

    def serialize(self) -> bytes:
        """
        Serializes the section into a bytearray.
        """
        raise NotImplementedError

    def deserialize(self, data: bytes):
        """
        deserializes the section into a bytearray.
        """
        raise NotImplementedError



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
This class returns the hash algorithm and signature size for the given SoC.

This should be considered deprecated given that there are no plans to support hashes other than
sha256, and boards other than the mt3620...
"""

from hashlib import sha256, sha384


class SigningType:
    """
    Signing type enum
    """

    Invalid = 0
    ECDsa256 = 1
    ECDsa384 = 2


class SocType:
    """
    SoC type enum
    """

    mt3620 = 0
    conundrum = 0x1
    conundrumv8 = 0x2
    conundrumr5 = 0x3
    rpi4 = 0x4
    imx7ulp = 0x5
    imx8ulp = 0x6


class SigningInfo:
    """
    A signing factory for the different SoCs that returns the correct
    signature algorithm and metadata.

    The signature is appended to the metadata section at the end of the file.
    Signing is performed over the metadata and image data combined (i.e. from byte 0:end).
    """

    def _signing_info_type(self):
        """
        Returns the signing type for the given SoC
        """
        if self.soc == SocType.mt3620:
            return SigningType.ECDsa256

        if self.soc == SocType.conundrum:
            return SigningType.ECDsa384

        if self.soc == SocType.conundrumv8:
            return SigningType.ECDsa384

        if self.soc == SocType.conundrumr5:
            return SigningType.ECDsa384

        if self.soc == SocType.rpi4:
            return SigningType.ECDsa256

        if self.soc == SocType.imx7ulp:
            return SigningType.ECDsa256

        if self.soc == SocType.imx8ulp:
            return SigningType.ECDsa384

        return SigningType.Invalid

    def __init__(self, soc_type=SocType.mt3620):
        self.soc = soc_type
        self.type = self._signing_info_type()

    def get_signature_size(self):
        """
        Returns the size of the signature in bytes.
        """
        if self.type == SigningType.ECDsa256:
            return 64

        if self.type == SigningType.ECDsa384:
            return 96

        return 64

    def get_hash_algorithm(self):
        """
        Returns the hash algorithm for the given SoC.
        """
        if self.type == SigningType.ECDsa256:
            return sha256

        if self.type == SigningType.ECDsa384:
            return sha384

        raise Exception("Invalid signing type")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Documents the expected structure of the OneBlInformation section.
Like capability files, this is not a metadata section, but rather the content.
"""

from ..metadata import MetadataSection
import struct


class OneBlInformationSection(MetadataSection):

    """
    Documents the binary format of OneBlInformation:
        uint32_t Information1
        uint32_t Information2
        uint32_t Information3
        uint32_t ChipId
        uint32_t Major
        uint32_t Minor
        uint32_t Patch
        uint32_t SecurityVersionNumber
    """

    def __init__(self, data):
        super().__init__("OneBlInformation", None, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.information1 = 0
            self.information2 = 0
            self.information3 = 0
            self.chip_id = 0
            self.major = 0
            self.minor = 0
            self.patch = 0
            self.security_version_number = 0

    @staticmethod
    def size():
        """
        Returns the size of the section in bytes.
        """
        return 32

    def deserialize(self, data):
        """
        Deserializes the section from binary data.
        """
        self.raw_bytes = data
        self.information1 = struct.unpack_from("<I", data, 0)[0]
        self.information2 = struct.unpack_from("<I", data, 4)[0]
        self.information3 = struct.unpack_from("<I", data, 8)[0]
        self.chip_id = struct.unpack_from("<I", data, 12)[0]
        self.major = struct.unpack_from("<I", data, 16)[0]
        self.minor = struct.unpack_from("<I", data, 20)[0]
        self.patch = struct.unpack_from("<I", data, 24)[0]
        self.security_version_number = struct.unpack_from("<I", data, 28)[0]

    def serialize(self):
        """
        Serializes the section to binary data.
        """
        return struct.pack(
            "<IIIIIIII",
            self.information1,
            self.information2,
            self.information3,
            self.chip_id,
            self.major,
            self.minor,
            self.patch,
            self.security_version_number,
        )

    def __str__(self):
        out_lines = []
        out_lines.append(f"Security Version:  {self.security_version_number}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
This enum lists all of the possible metadata section IDs.

This enum is used by the MetadataSection class to identify the type of metadata section that is being parsed.
It is also reflectively used when parsing an image to determine which metadata sections are present in the image
and the class that should be invoked on the data.
"""


class MetadataSectionId:
    # /// <summary>
    # /// None / not defined
    # /// </summary>
    Undefined = 0

    # /// <summary>
    # /// 'AD' - Legacy ABI depends info (Processed it as 'unkknown')
    # /// </summary>
    LegacyABIDepends = 0x4441

    # /// <summary>
    # /// 'AP' - Legacy ABI provides info (Processed it as 'unkknown')
    # /// </summary>
    LegacyABIProvides = 0x5041

    # /// <summary>
    # /// 'ND' - ABI depends info
    # /// </summary>
    ABIDepends = 0x444E

    # /// <summary>
    # /// 'NP' - ABI provides info
    # /// </summary>
    ABIProvides = 0x504E

    # /// <summary>
    # /// 'CM' - Compression info
    # /// </summary>
    Compression = 0x4D43

    # /// <summary>
    # /// 'DB' - Debug info
    # /// </summary>
    Debug = 0x4244

    # /// <summary>
    # /// 'LG' - legacy footer format
    # /// </summary>
    Legacy = 0x474C

    # /// <summary>
    # /// 'ID' - Identity info
    # /// </summary>
    Identity = 0x4449

    # /// <summary>
    # /// 'RV' - Revocation info
    # /// </summary>
    Revocation = 0x5652

    # /// <summary>
    # /// 'SG' - signature info
    # /// </summary>
    Signature = 0x4753

    # /// <summary>
    # /// 'TP' Temporary image info
    # /// </summary>
    TemporaryImage = 0x5054

    # /// <summary>
    # /// 'RO' Required offset info
    # /// </summary>
    RequiredFlashOffset = 0x4F52

    # /// <summary>
    # /// 'IP' Image policy info
    # /// </summary>
    ImagePolicy = 0x5049



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The temporary image section describes a development image to be sideloaded to the device.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct


class TempImageMetadataFlags:
    """
    Temporary image metadata flags.
    """

    RemoveAtBoot = 0x1
    UnderDevelopment = 0x2

    def __init__(self, data=None):
        if data is not None:
            self.deserialize(data)
        else:
            self.flags = 0

    def deserialize(self, data):
        """
        Deserializes the flags from the given data.
        """
        self.flags = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        """
        Serializes the flags to bytes.
        """
        return struct.pack("<I", self.flags)

    def get_flag_names(self):
        """
        Converts the flags to a string array.
        """
        names = []
        if self.flags & TempImageMetadataFlags.RemoveAtBoot:
            names += ["RemoveAtBoot"]
        elif self.flags & TempImageMetadataFlags.UnderDevelopment:
            names += ["UnderDevelopment"]
        return names

    def __str__(self):
        return ", ".join(self.get_flag_names())


class TempImageMetadataSection(MetadataSection):
    """
    The temporary image section describes the properties a development image to be sideloaded to the device.
    """

    # TempImageMetadata
    # uint32_t flags

    def __init__(self, data=None):
        super().__init__("Temporary Image", MetadataSectionId.TemporaryImage, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.flags = TempImageMetadataFlags(None)

    @staticmethod
    def size():
        """
        Returns the size of the temporary image section in bytes.
        """
        return 4

    def deserialize(self, data):
        """
        Deserializes the temporary image section from the given data.
        """
        self.flags = TempImageMetadataFlags(data)

    def serialize(self):
        """
        Serializes the temporary image section to bytes.
        """
        return self.flags.serialize()

    def __str__(self):
        remove_at_boot = "RemoveAtBoot" in self.flags.get_flag_names()
        under_development = "UnderDevelopment" in self.flags.get_flag_names()

        out_lines = []
        out_lines.append(f"Remove image at boot: {remove_at_boot}")
        out_lines.append(f"Under development: {under_development}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Compression section, if present, shows what compression algorithm was used to compress the image.
It also gives the uncompressed size of the image.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct


class CompressionType:
    """
    Compression type enum.
    """

    Invalid = 0
    Lzma = 1

    compression_type: int

    def __init__(self, data):
        if data is not None:
            self.deserialize(data)
        else:
            self.compression_type = 0

    @staticmethod
    def size():
        """
        Returns the size of the compression type in bytes.
        """
        return 4

    def get_compression_type_name(self):
        """
        Converts the compression type to a string
        """
        if self.compression_type == CompressionType.Lzma:
            return "Lzma"

        raise Exception(
            "CompressionSection: Invalid compression type %d" % (self.compression_type)
        )

    def deserialize(self, data):
        """
        Deserializes the compression type from the given data.
        """
        self.compression_type = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        return struct.pack("<I", self.compression_type)

    def __str__(self):
        return str(self.compression_type)


class CompressionSection(MetadataSection):
    """
    The compression section, if present, shows what compression algorithm was used to compress the image.
    """

    # uint32_t uncompressed_size
    # uint32_t compression_type

    uncompressed_size: int
    compression_type: CompressionType

    def __init__(self, data=None):
        super().__init__("Compression", MetadataSectionId.Compression, data)

        if data is not None:
            self.deserialize(data)
        else:
            self.uncompressed_size = 0
            self.compression_type = CompressionType(None)

    @staticmethod
    def size():
        """
        Returns the size of the compression section in bytes.
        Uncompressed member size + compression type size.
        """
        return CompressionType.size() + 4

    def deserialize(self, data):
        """
        Deserializes the compression section from the given data.
        """
        self.uncompressed_size = struct.unpack_from("<I", data, 0)[0]
        self.compression_type = CompressionType(data[4:])

    def serialize(self):
        return (
            struct.pack("<I", self.uncompressed_size)
            + self.compression_type.serialize()
        )

    def __str__(self):
        out_lines = []
        out_lines.append(f"Type:              {self.compression_type}")
        out_lines.append(f"Uncompressed size: {self.uncompressed_size:,} bytes")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Contains definition for the different headers used in image metadata.
"""

import struct


class ImageMetadataHeader:
    """
    image metadata header
    describes the layout of the image metadata (i.e. number of sections).

    contains:
    uint32_t magic 0x4d345834
    uint32_t section_count
    """

    CorrectMagicValue = 0x4D345834

    def __init__(self, data=None):
        if data is not None:
            self.deserialize(data)
        else:
            self.magic_value = self.CorrectMagicValue
            self.section_count = 0

    @staticmethod
    def size() -> int:
        """
        size of the header (8 bytes).
        """
        return 8

    def deserialize(self, data: bytearray):
        """
        deserialize the header and check magic values match.
        """
        self.magic_value = struct.unpack_from("<I", data, 0)[0]
        self.section_count = struct.unpack_from("<I", data, 4)[0]

        if self.magic_value != ImageMetadataHeader.CorrectMagicValue:
            raise Exception("ImageMetadataHeader: Invalid magic value")

    def serialize(self) -> bytearray:
        """
        serialize the header.
        """
        return struct.pack("<II", self.magic_value, self.section_count)

    def __str__(self):
        return (
            "ImageMetadataHeader\n\tmagic_value: "
            + str(self.magic_value)
            + "\n\tsection_count: "
            + str(self.section_count)
            + "\n"
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The ABI depends/provides list software dependencies/requirements of a given image.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct
from typing import List


class ABIDependsType:
    """
    The type of dependency.
    """

    NoDependency = 0
    SecureWorldRuntime = 1
    OSRuntime = 2
    ApplicationRuntime = 3


class ABIIdentity:
    """
    ABI Identity wraps the AbiDependsType.
    It provides deserialisation and stringification.
    """

    # uint32_t version
    # uint32_t type

    def __init__(self, data):
        self.deserialize(data)

    @staticmethod
    def size():
        """
        The size of the ABI Identity in bytes.
        """
        return 8

    def deserialize(self, data):
        """
        Deserialise the ABI Identity member properties from the given data.
        """
        self.raw_bytes = data
        self.version = struct.unpack_from("<I", data, 0)[0]
        self.type = struct.unpack_from("<I", data, 4)[0]

    def serialize(self):
        return struct.pack("<II", self.version, self.type)

    def get_type_name(self):
        """
        Get the stringified name of the ABI Identity type.
        """
        if self.type == ABIDependsType.NoDependency:
            return "NoDependency"
        elif self.type == ABIDependsType.SecureWorldRuntime:
            return "SecureWorldRuntime"
        elif self.type == ABIDependsType.OSRuntime:
            return "OSRuntime"
        elif self.type == ABIDependsType.ApplicationRuntime:
            return "ApplicationRuntime"
        else:
            raise Exception("Unknown AbiDependsIdentity type: %d" % (self.type))

    def __str__(self):
        return f"{self.get_type_name()}@{self.version}"


class ABISection(MetadataSection):
    """
    The superclass of ABIProvides/ABIDepends.
    It provides functionality common to both.
    """

    versions: List[ABIIdentity]
    version_count: int

    def __init__(self, name, section_id, data):
        super().__init__(name, section_id, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.version_count = 0
            self.versions = []

    @staticmethod
    def size():
        """
        The size of the ABI section in bytes.
        """
        return 4

    def deserialize(self, data):
        """
        Deserialise the ABI section member properties from the given data.
        """
        self.version_count = struct.unpack_from("<I", data, 0)[0]
        self.versions = []
        offset = 4
        for _ in range(self.version_count):
            self.versions += [ABIIdentity(data[offset : offset + ABIIdentity.size()])]
            offset += ABIIdentity.size()

    def serialize(self):
        """
        Serialise the ABI section member properties.
        """
        out = struct.pack("<I", len(self.versions))
        for v in self.versions:
            out += v.serialize()
        return out

    def _to_string(self, prefix):
        """
        Helper function to stringify the ABI section.
        """
        out_lines = []
        for v in self.versions:
            out_lines.append(f"{prefix}{v}")
        return "\n".join(out_lines)


class ABIProvidesSection(ABISection):
    """
    The ABI Provides section lists any binary dependencies that the image
    provides.
    """

    def __init__(self, data=None):
        super().__init__("ABI Provides", MetadataSectionId.ABIProvides, data)

    def __str__(self):
        return self._to_string("Provides:          ")


class ABIDependsSection(ABISection):
    """
    The ABI Depends section lists any binary dependencies that the image
    requires.
    """

    def __init__(self, data=None):
        super().__init__("ABI Depends", MetadataSectionId.ABIDepends, data)

    def __str__(self):
        return self._to_string("Depends:           ")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
For sections that are not expected to be seen (i.e. legacy data structures before extensible metadata),
the not implemented section is used.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection


class NotImplementedSection(MetadataSection):
    """
    NotImplementedSection is used for sections that are not expected to be seen.
    """

    def __init__(self, data):
        super().__init__("NotImplemented", MetadataSectionId.Undefined, data)
        self.data = data

    def __str__(self):
        return "NotImplementedSection\n\tdata: %s" % (self.data) + "\n"



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The revocation section, iff present, an image can be revoked using fused based revocation.
This revocation check is similar to the ROM revocation of 1BLs and involves checking a security version
number against a 32 bit SW fuse.  If bit N is set to 1 then security version N is not trusted.
In general this version would only be increased when we need to do revocation and is otherwise static
 from build to build.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct


class RevocationSection(MetadataSection):
    """
    The class that represent the revocation section.
    """

    name = "Revocation"
    section_id = MetadataSectionId.Revocation

    # uint32_t version
    def __init__(self, data=None):
        super().__init__("Revocation", MetadataSectionId.Revocation, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.version = 0

    @staticmethod
    def size():
        """
        Returns the size of the revocation section in bytes.
        """
        return 4

    def deserialize(self, data):
        """
        Deserializes the revocation section from the given data.
        """
        self.version = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        """
        Serializes the revocation section to bytes.
        """
        return struct.pack("<I", self.version)

    def __str__(self):
        out_lines = []
        out_lines.append(f"Security Version:  {self.version}")
        return "\n".join(out_lines)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

import struct
from ..metadata_sections.utils import name_from_section_id


class ImageMetadataSectionHeader:
    """
    section metadata header describes the upcoming section.
    uint16_t section_id
    uint16 length
    bytes[length]
    """

    def __init__(self, data=None):
        if data is not None:
            self.deserialize(data)
        else:
            self.section_id = 0
            self.data_length = 0

    @staticmethod
    def size() -> int:
        """
        size of the section header (4 bytes).
        """
        return 4

    def deserialize(self, data: bytearray):
        """
        deserialize the section header from bytes.
        """
        self.section_id = struct.unpack_from("<H", data, 0)[0]
        self.data_length = struct.unpack_from("<H", data, 2)[0]

    def serialize(self) -> bytearray:
        """
        serialize the section header into bytes.
        """
        return struct.pack("<HH", self.section_id, self.data_length)

    def __str__(self) -> str:
        return (
            "ImageMetadataSectionHeader\n\t"
            + f"section_id: {self.section_id:x}"
            + f"({name_from_section_id(self.section_id)})\n\t"
            + f"data_length: {self.data_length}"
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Takes a signature stored in the image and converts it to DER encoded format for use with pyca/cryptography.
C# code simply appends the R/S consecutively in byte format for the signature. This is not DER encoded.
"""

import struct


class DEREncodedSignature:
    @staticmethod
    def from_stored_signature(stored_signature: bytes) -> bytes:
        """
        Converts a stored signature to DER encoded format.
        """
        # C#/dotnet does not output a DER encoded signature, so we need to convert it.
        # der encoded signature is generally:
        # 0x30, payload length, 0x20, r length, r, 0x20, s length, s
        # for compatibility with dotnet, r length and s length must be 32 bytes long
        # if the first byte of r/s is >= 0x80, then 0 is added to the front of the r/s.
        # if the first byte of r/s is 0x00, then it is removed.
        # This is the only format supported.
        if len(stored_signature) != 64:
            raise Exception("Invalid stored signature length")
        r = stored_signature[:32]
        s = stored_signature[32:]

        to_encode = [r, s]
        der_encoded = bytearray([0x30])

        encoded_buf = bytearray()
        for sig in to_encode:
            if sig[0] >= 0x80:
                encoded_buf.extend(bytearray([0x02, len(sig) + 1, 0]))
            else:
                encoded_buf.extend(bytearray([0x02, len(sig)]))
            encoded_buf.extend(sig)

        der_encoded.extend([len(encoded_buf)])
        der_encoded.extend(encoded_buf)
        return bytes(der_encoded)

    @staticmethod
    def to_stored_signature(der_encoded_signature: bytes) -> bytes:
        """
        Converts a DER encoded signature to the format used by the Azure Sphere SDK.
        This code does not support all DER encoded signatures, only the format used by the Azure Sphere SDK.
        """
        # der encoded signature is generally:
        # 0x30, payload length, 0x20, r length, r, 0x20, s length, s
        # for compatibility with dotnet, r length and s length must be 32 bytes long
        # if the first byte of r/s is >= 0x80, then 0 is added to the front of the r/s.
        # if the first byte of r/s is 0x00, then it is removed

        # extract the DER signature header
        der_type, payload_length = struct.unpack_from("BB", der_encoded_signature)
        if der_type != 0x30:
            raise Exception("Invalid DER encoded signature. Incorrect type.")

        if payload_length != len(der_encoded_signature) - 2:
            raise Exception("Invalid DER encoded signature. Incorrect payload length.")

        remainder = der_encoded_signature[2:]
        signature_values = []  # r/s value storage
        while remainder != bytearray():
            (var_type, sig_length) = struct.unpack_from("BB", remainder)
            remainder = remainder[2:]  # consume the type and length info.
            # validate type as integer
            if var_type != 0x02:
                raise Exception("Invalid DER encoded signature")
            # ensure length is 32 bytes or 33 bytes (if first byte is 0x00)
            if sig_length not in [0x20, 0x21]:
                raise Exception("Invalid DER encoded signature")

            if sig_length == 0x21:
                if remainder[0] != 0x00:
                    raise Exception("Invalid DER encoded signature")
                # consume the byte
                remainder = remainder[1:]
            # stored the signature value
            signature_values.append(remainder[:sig_length])
            remainder = remainder[sig_length:]

        ## return r combined with s
        return signature_values[0] + signature_values[1]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
metadata utility functions to retrieve a name from a section id and to retrieve a class from a section id.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
from .abi import ABIDependsSection, ABIProvidesSection
from .compression import CompressionSection
from .debug import DebugSection
from .identity import IdentitySection
from .image_policy import ImagePolicySection
from .required_offset import RequiredOffsetSection
from .revocation import RevocationSection
from .signature import SignatureSection
from .temporary_image import TempImageMetadataSection
from .not_implemented import NotImplementedSection


def name_from_section_id(section_id: MetadataSectionId) -> str:
    """
    Returns the name of the section from the section id.
    """
    if section_id == MetadataSectionId.Undefined:
        return "Undefined"
    elif section_id == MetadataSectionId.LegacyABIDepends:
        return "LegacyABIDepends"
    elif section_id == MetadataSectionId.LegacyABIProvides:
        return "LegacyABIProvides"
    elif section_id == MetadataSectionId.ABIDepends:
        return "ABIDepends"
    elif section_id == MetadataSectionId.ABIProvides:
        return "ABIProvides"
    elif section_id == MetadataSectionId.Compression:
        return "Compression"
    elif section_id == MetadataSectionId.Debug:
        return "Debug"
    elif section_id == MetadataSectionId.Legacy:
        return "Legacy"
    elif section_id == MetadataSectionId.Identity:
        return "Identity"
    elif section_id == MetadataSectionId.Revocation:
        return "Revocation"
    elif section_id == MetadataSectionId.Signature:
        return "Signature"
    elif section_id == MetadataSectionId.TemporaryImage:
        return "TemporaryImage"
    elif section_id == MetadataSectionId.RequiredFlashOffset:
        return "RequiredFlashOffset"
    elif section_id == MetadataSectionId.ImagePolicy:
        return "ImagePolicy"
    else:
        return "Unknown"


def class_from_section_id(section_id: MetadataSectionId) -> MetadataSection:
    """
    Metadata section factory that returns the class of the section from the section id.
    """
    if section_id == MetadataSectionId.Undefined:
        return NotImplementedSection
    elif section_id == MetadataSectionId.LegacyABIDepends:
        return NotImplementedSection
    elif section_id == MetadataSectionId.LegacyABIProvides:
        return NotImplementedSection
    elif section_id == MetadataSectionId.ABIDepends:
        return ABIDependsSection
    elif section_id == MetadataSectionId.ABIProvides:
        return ABIProvidesSection
    elif section_id == MetadataSectionId.Compression:
        return CompressionSection
    elif section_id == MetadataSectionId.Debug:
        return DebugSection
    elif section_id == MetadataSectionId.Legacy:
        return NotImplementedSection
    elif section_id == MetadataSectionId.Identity:
        return IdentitySection
    elif section_id == MetadataSectionId.Revocation:
        return RevocationSection
    elif section_id == MetadataSectionId.Signature:
        return SignatureSection
    elif section_id == MetadataSectionId.TemporaryImage:
        return TempImageMetadataSection
    elif section_id == MetadataSectionId.RequiredFlashOffset:
        return RequiredOffsetSection
    elif section_id == MetadataSectionId.ImagePolicy:
        return ImagePolicySection
    else:
        raise Exception("Unknown section id: " + str(section_id))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

from .image_metadata import ImageMetadata
from .signing.app_development import AppDevelopmentSigner
from .signing.der_encoded import DEREncodedSignature


class Image:
    def __init__(self, data=None):
        self.signature = None
        self.data = data
        self.metadata = ImageMetadata(None)
        if data is not None:
            self.deserialize(data)

    def as_bytes(self, signer=AppDevelopmentSigner(), include_signature=False):
        new_data = self.data + self.metadata.serialize()
        if include_signature:
            if self.signature is None:
                self.signature = DEREncodedSignature.to_stored_signature(
                    signer.sign_data(new_data)
                )
            return new_data + self.signature
        return new_data

    def deserialize(self, data):
        try:
            self.metadata = ImageMetadata(data)
        except Exception as e:
            print(f"Metadata parsing failed: {e}")
            # metadata parsing could not correctly find the metadata
            return

        self.data = data[: self.metadata.start_of_metadata]
        self.signature = data[self.metadata.end_of_metadata :]

        if len(self.signature) != self.metadata.signature_size:
            raise Exception("Signature size mismatch")

    def serialize(self, signer=AppDevelopmentSigner(), include_signature=True):
        return self.as_bytes(signer, include_signature)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
The required offset section (from what can be gleaned from code) shows the offset
 in flash where the image must be placed.
"""

from .metadata_id import MetadataSectionId
from .metadata import MetadataSection
import struct


class RequiredOffsetSection(MetadataSection):
    """
    The required offset section (from what can be gleaned from code) shows the offset
    in flash where the image must be placed.
    """

    offset: int
    # uint32_t offset

    def __init__(self, data=None):
        super().__init__("Required Offset", MetadataSectionId.RequiredFlashOffset, data)
        if data is not None:
            self.deserialize(data)
        else:
            self.offset = 0

    @staticmethod
    def size():
        """
        Returns the size of the required offset section in bytes.
        """
        return 4

    def deserialize(self, data):
        """
        Deserializes the required offset section from the given data.
        """
        self.offset = struct.unpack_from("<I", data, 0)[0]

    def serialize(self):
        """
        Serializes the required offset section to bytes.
        """
        return struct.pack("<I", self.offset)

    def __str__(self):
        out_lines = []
        out_lines.append(f"Offset: {self.offset:#x}")
        return "\n".join(out_lines)
